//: $CC -c $file $OPTIONS
#ifndef SP_SMATRIXC
#define SP_SMATRIXC
#include "sp_names.h"
#include SP_SMATRIX_H

const sp_sMatrix& sp_sMatrix::operator=(const sp_sMatrix &B)
{ sp_listU<ls_REAL>::operator=(B); return *this;}

sp_sMatrix sp_sMatrix::operator+(const sp_sMatrix &B)const
{ static char *q1="sp_sMatrix::operator+";
  ls_UINT n=B.dimension();
  if(n!=a->j) throw ls_error(name,q1,size_error_);
  sp_sMatrix C(n);
  list2<ls_REAL> *b=B.asList(), *aa=a->r_next, *bb=b->r_next, *an, *bn;
  ls_UINT i;
  for(i=0; i<n; i++, aa++, bb++)
  { an=aa->r_next; bn=bb->r_next;
    while(an!=a)
      if(an->j<bn->j) C.put(an->x,i,an->j), an=an->r_next;
      else
        if(an->j>bn->j) C.put(bn->x,i,bn->j), bn=bn->r_next;
        else C.put(an->x+bn->x,i,an->j), an=an->r_next, bn=bn->r_next;
    while(bn!=b) C.put(bn->x,i,bn->j), bn=bn->r_next;
  }
  return C;
}

sp_sMatrix sp_sMatrix::operator-(const sp_sMatrix &B)const
{ static char *q1="sp_sMatrix::operator-";
  ls_UINT n=B.dimension();
  if(n!=a->j) throw ls_error(name,q1,size_error_);
  sp_sMatrix C(n);
  list2<ls_REAL> *b=B.asList(), *aa=a->r_next, *bb=b->r_next, *an, *bn;
  ls_UINT i;
  for(i=0; i<n; i++, aa++, bb++)
  { an=aa->r_next; bn=bb->r_next;
    while(an!=a)
      if(an->j<bn->j) C.put(an->x,i,an->j), an=an->r_next;
      else
        if(an->j>bn->j) C.put(bn->x,i,bn->j), bn=bn->r_next;
        else C.put(an->x-bn->x,i,an->j), an=an->r_next, bn=bn->r_next;
    while(bn!=b) C.put(bn->x,i,bn->j), bn=bn->r_next;
  }
  return C;
}

sp_sMatrix& sp_sMatrix::operator+=(const sp_sMatrix &B)
{ static char *q1="sp_sMatrix::operator+=";
  ls_UINT n=B.dimension();
  if(n!=a->j) throw ls_error(name,q1,size_error_);
  sp_sMatrix C(n);
  list2<ls_REAL> *b=B.asList(), *aa=a->r_next, *bb=b->r_next, *an, *bn;
  ls_UINT i;
  for(i=0; i<n; i++, aa++, bb++)
  { an=aa->r_next; bn=bb->r_next;
    while(an!=a)
      if(an->j<bn->j) C.put(an->x,i,an->j), an=an->r_next;
      else
        if(an->j>bn->j) C.put(bn->x,i,bn->j), bn=bn->r_next;
        else C.put(an->x+bn->x,i,an->j), an=an->r_next, bn=bn->r_next;
    while(bn!=b) C.put(bn->x,i,bn->j), bn=bn->r_next;
  }
  swap(C); C.remove();
  return *this;
}

sp_sMatrix& sp_sMatrix::operator-=(const sp_sMatrix &B)
{ static char *q1="sp_sMatrix::operator-=";
  ls_UINT n=B.dimension();
  if(n!=a->j) throw ls_error(name,q1,size_error_);
  sp_sMatrix C(n);
  list2<ls_REAL> *b=B.asList(), *aa=a->r_next, *bb=b->r_next, *an, *bn;
  ls_UINT i;
  for(i=0; i<n; i++, aa++, bb++)
  { an=aa->r_next; bn=bb->r_next;
    while(an!=a)
      if(an->j<bn->j) C.put(an->x,i,an->j), an=an->r_next;
      else
        if(an->j>bn->j) C.put(bn->x,i,bn->j), bn=bn->r_next;
        else C.put(an->x-bn->x,i,an->j), an=an->r_next, bn=bn->r_next;
    while(bn!=b) C.put(bn->x,i,bn->j), bn=bn->r_next;
  }
  swap(C); C.remove();
  return *this;
}

sp_Vector sp_sMatrix::operator* (const sp_Vector &v) const
{ static char *q1="sp_sMatrix::operator*";
  ls_UINT n=a->j;
  if(v.dimension()!=n) throw ls_error(v.name,q1, size_error_);
  ls_UINT i, j;
  list2<ls_REAL> *aa, *anf=a->r_next;
  sp_Vector u(n);
  double s, sp, sm;
  list1<ls_REAL> *va=v.asList(), *vp, *vv, *ua=u.asList(), *uu=ua;
  for (i=0; i<n; i++)
  { sp=sm=0, aa=anf+i, vp=va;
    while((aa=aa->c_next)->i<i)
    { j=aa->i; while((vv=vp->next)->i<j) vp=vv;
      if(vv->i==j) if((s=aa->x * vv->x)>0) sp+=s; else sm+=s;
      if(vv==va) break;
    }
    aa=a+i;
    while((aa=aa->r_next)!=a)
    { j=aa->j; while((vv=vp->next)->i<j) vp=vv;
      if(vv->i==j) if((s=aa->x * vv->x)>0) sp+=s; else sm+=s;
      if(vv==va) break;
    }
    s=sp+sm;
    if(s)
    { uu=uu->next=u.get1_el(); uu->i=i, uu->x=ls_REAL(s), uu->next=ua;}
  }
  return u;
}

ls_Vector sp_sMatrix::operator* (const ls_Vector &v) const
{ static char *q1="sp_sMatrix::operator*";
  ls_UINT n=a->j;
  if(v.dimension()!=n) throw ls_error(v.name,q1, size_error_);
  ls_UINT i;
  list2<ls_REAL> *aa, *anf=a->r_next;
  ls_Vector u(n);
  ls_REAL *va=v.asArray(), *uu=u.asArray();
  double s, sp, sm;
  for (i=0; i<n; uu[i]=ls_REAL(sp+sm), i++)
  { sp=sm=0, aa=anf+i;
    while((aa=aa->c_next)->i<i)
      if((s=va[aa->i]*aa->x)>0) sp+=s; else sm+=s;
    aa=anf+i;
    while((aa=aa->r_next)!=a)
      if((s=va[aa->j]*aa->x)>0) sp+=s; else sm+=s;
  }
  return u;
}

sp_Matrix sp_sMatrix::operator*(const sp_sMatrix &B)const
{ static char *q1="sp_sMatrix::*";
  ls_UINT n=B.dimension();
  if(n!=a->j) throw ls_error(name,q1,size_error_);
  sp_Matrix C(n,n);
  ls_Vector v(n);
  ls_REAL *vv=v.asArray();
  list2<ls_REAL>* b=B.asList(), *anf=b->r_next, *aa;
  for(ls_UINT i=0; i<n; i++)
  { aa=anf+i; while((aa=aa->r_next)!=b) vv[aa->j]=aa->x;
    aa=anf+i; while((aa=aa->c_next)!=b) vv[aa->i]=aa->x;
    C.put_column((*this)*v,0,i);
    memset(vv,0,sizeof(ls_REAL)*n);
  }
  return C;
}

ls_UINT sp_sMatrix::solve(ls_Vector &x, ls_Vector &b) const
{ ls_UINT n=a->j;
  ls_Vector r, p, w; r=b-(*this)*x; p=r; w=(*this)*p;
  ls_REAL s, rr=r*r, wp=w*p, rrn, eps=x.eps;
  ls_UINT it=0, rep=5*n;
  while(!((rr+wp)<eps)&&(it++<rep))
  { s=rr/wp; r-=s*w; x+=s*p; rrn=r*r; p=r+(rrn/rr)*p;
    w=(*this)*p; wp=w*p; rr=rrn;
  }
  b=r; return it;
}
#endif
